package lazabs.types

import lazabs.ast.ASTree._
import lazabs.types._

// performs basic type checking for the moment; just assign types to variables according to the environment

object TypeCheck{
  /**
   * The environment at each point of the program execution
   */
  var environ = Map[String,Type]().empty

  /**
   * extracts the global environment of a function
   */
  def globalEnviron(so: Sobject): Map[String,Type] = so.defs match {
    case Nil => Map.empty
    case VarDeclaration(name, t, value) :: ds => 
      val environ = globalEnviron(Sobject(so.name, ds))
      environ + (name -> t)
    case _ :: ds => globalEnviron(Sobject(so.name, ds))  
  }
  
  def apply(so: Sobject): Sobject = {
    val global = globalEnviron(so)
    Sobject(so.name, typeCheck(so.defs,global).asInstanceOf[List[Declaration]])
  }
  
  def typeCheck(declList : List[ASTree], env: Map[String,Type]): List[ASTree] = declList match {
    case Nil => Nil
    case VarDeclaration(name, t, value) :: rest =>
      VarDeclaration(name, t, typeCheck(value,env)) :: typeCheck(rest, env + (name -> t))
    case FunctionDefinition(funcName, ps, t, e) :: rest =>
      val funcEnv = env ++ (ps.map(p => (p.name, p.typ)))
      val typeCheckedBody = typeCheck(e, funcEnv)
      FunctionDefinition(funcName, ps, t, typeCheckedBody).stype(t) :: typeCheck(rest, env)
    case (FunctionCall("sc_havoc", List(v@Variable(vname,_)))) :: rest if(v.stype != UnitType()) =>
      declList.head :: typeCheck(rest, env + (vname -> v.stype))
    case (e: Expression) :: rest  =>
      typeCheck(e, env) :: typeCheck(rest, env)
    case _ :: rest =>
      declList.head :: typeCheck(rest, env)
  }

  def typeCheck(e: Expression, env: Map[String,Type]): Expression = e match {
    case Block(declList) => Block(typeCheck(declList, env))
    case FunctionCall(funcName, exprList) =>
      FunctionCall(funcName, exprList.map(e => typeCheck(e,env)))
    case IfThenElse(cond, conseq, altern) => IfThenElse(typeCheck(cond, env), typeCheck(conseq, env), typeCheck(altern, env))
    case WhileLoop(cond, body) => WhileLoop(typeCheck(cond, env), typeCheck(body, env))
    case DoWhileLoop(cond, body) => DoWhileLoop(typeCheck(cond, env), typeCheck(body, env))    
    case Universal(v, qe) => Universal(v, typeCheck(qe, env + (v.name -> v.stype))).stype(e.stype)
    case Existential(v, qe) => Existential(v, typeCheck(qe, env + (v.name -> v.stype))).stype(e.stype)
    case CreateObject(cName, cArgs) => CreateObject(cName, cArgs.map(e => typeCheck(e,env))).stype(e.stype)
    case BinaryExpression(e1, op, e2) => 
      var te1 = typeCheck(e1, env)
      var te2 = typeCheck(e2, env)
      if((te1.stype == UnitType()) && (te2.stype != UnitType())) te1 = te1.stype(te2.stype)
      if((te1.stype != UnitType()) && (te2.stype == UnitType())) te2 = te2.stype(te1.stype)
      BinaryExpression(te1, op, te2).stype(te1.stype)
    case TernaryExpression(op, e1, e2, e3) => TernaryExpression(op, typeCheck(e1, env), typeCheck(e2, env), typeCheck(e3, env)).stype(e.stype)
    case Variable("sc_skip",None) => e.stype(UnitType())
    case Variable(name,d) if (e.stype == UnitType()) => Variable(name,d).stype(env.get(name) match {
      case Some(typ) => typ
      case None => println("Type error: type of variable " + name + " is unknown in the environment " + env); UnitType()
    })
    case UnaryExpression(op, e) => UnaryExpression(op, typeCheck(e, env))
    case NumericalConst(n) => e.stype(IntegerType())
    case _ =>
      e
  }
}